/* undo.c
 *
 * set of procedures to realize Undo and Redo edit operations;
 * maximum number of changes that may be undone is set by EditSTACK_SIZE
 *
 * Last revision - 18.01.1993
 *
 *
 * (C) COPYRIGHT A. Stochniol, 1991 - 1993
 * All Rights Reserved
 *
 * Author:
 * Dr. Andrzej Stochniol
 * Department of Mechanical Engineering
 * Imperial College of Science, Technology and Medicine
 * London, SW7 2BX, UK
 *
 * E-mail:  A.Stochniol@ic.ac.uk
 *
 *

 ****

   modified by: p.hlstede	(peter@calvados.chemie.uni-bremen.de)

   - multiple insertion or deletion actions are merged to single
     undo actions if possible
   - a function 'clear_undo_redo' is provided to allow for the
     reuse of redo/undo-actions e.g. after a file was saved
   - when text is deleted, only the deleted part of the text is
     retrieved from the widget (with XmTexGetSubString)

							06.09.1997
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <Xm/Xm.h>
#include <Xm/Text.h>

#include "sys_consts.h"
#include "globals.h"
#include "utils.h"
#include "undo.h"
#include "callbacks.h"
#include "dstring.h"


/* size of the undo/redo stacks */
#define EditSTACK_SIZE		100


typedef struct
{
   XmTextPosition   startPos, endPos;
   XmTextPosition   cursorPos;
   char            *del_text, *ins_text;
   int              step;
}
EditAction, *EditActionPtr;

typedef struct
{
   int      bottom;		/* bootom (head) of the stack */
   int      top;		/* top (tail) of the stack */
   EditActionPtr el[EditSTACK_SIZE];	/* pointers to stack elements */
}
EditStack;


/***  PROCEDURE DECLARATIONS  ***/
static void flush_edit_stack(EditStack * s);
static void pop_edit_stack(EditStack * s, EditActionPtr * el, Boolean dec);
static void push_edit_stack(EditStack * s, EditActionPtr el, Boolean inc);


/***  GLOBAL VARIABLES  ***/
Boolean  undo_redo_action = False;


/***  LOCAL VARIABLES  ***/
/* to speed the process of editing we preallocate a buffer for edit_history,
   with exactly the same length as EditSTACK_SIZE; this get rid of permanent
   allocation and deallocation of memory for every edit action; the memory
   must be only allocated/deallocated to store text added (not for all
   variables describing the edit operation) */
static EditAction edit_history[EditSTACK_SIZE];
static Boolean edit_history_initialized = False;

static EditStack undo_stack = { 0, 0 };
static EditStack redo_stack = { 0, 0 };

static char *WordSeps = " \t\n";


static void push_edit_stack(EditStack *s, EditActionPtr el, Boolean inc)
{
   /* push an element on the edit stack (on its top); the stack has predefined
      maximum length (EditSTACK_SIZE); if the stack is filled up one element
      from the bottom of the stack is destroyed, address of the bottom is moved
      up so we get one space on the stack. In fact it is not a classic
      stack which I implemented here but this "cyclic" version gives an easy
      way of restoring last EditSTACK_SIZE edit commands;
      because of the possibility of moving of the stack bottom the address
      of any element on the stack must be calculated "modulo" EditSTACK_SIZE;
    */

   s->el[s->top] = el;
   s->top = (s->top + 1) % EditSTACK_SIZE;

   /* check if the stack is used up completely; if so destroy the oldest
      element (i.e. from the bottom of the stack) and move the address
      of the bottom
    */
   if (s->bottom == s->top)
   {
      /* if memory was allocated for the text free it */
      if (s->el[s->bottom]->del_text)
      {
	 XtFree(s->el[s->bottom]->del_text);
	 s->el[s->bottom]->del_text = NULL;
      }
      if (s->el[s->bottom]->ins_text)
      {
	 XtFree(s->el[s->bottom]->ins_text);
	 s->el[s->bottom]->ins_text = NULL;
      }

      s->bottom = (s->bottom + 1) % EditSTACK_SIZE;
   }
   if (inc == True)
      changes_counter += el->step;
}				/* push_edit_stack */


/* pop the element from the top of the edit stack s */
static void  pop_edit_stack(EditStack *s, EditActionPtr *el, Boolean dec)
{
   if (s->bottom != s->top)
   {
      s->top--;
      if (s->top < 0)
	 s->top += EditSTACK_SIZE;	/* "cyclic" stack */
      *el = s->el[s->top];	/* moved here to avoid
				   checking twice for sign
				   of s->top -- */
      if (dec == True)
	 changes_counter -= (*el)->step;
   }
   else
      *el = NULL;		/* empty stack */
}				/* pop_edit_stack */


/* flush the edit stack pointed by s */
static void  flush_edit_stack(EditStack *s)
{
   int      i, j, n;
   n = s->top - s->bottom;
   if (n < 0)
      n += EditSTACK_SIZE;	/* "cyclic" stack */

   for (i = 0; i < n; i++)
   {
      j = s->top - i - 1;
      if (j < 0)
	 j += EditSTACK_SIZE;
      /* if memory was allocated for the text free it */
      if (s->el[j]->del_text)
      {
	 XtFree(s->el[j]->del_text);
	 s->el[j]->del_text = NULL;
      }

      if (s->el[j]->ins_text)
      {
	 XtFree(s->el[j]->ins_text);
	 s->el[j]->ins_text = NULL;
      }
   }
   s->top = s->bottom;
}				/* flush_edit_stack */


void  reset_undo_redo(void)
{
   int      i;

   if (edit_history_initialized == False)
      return;

   /* resets undo/redo by flushing redo and undo stacks */
   undo_stack.top = undo_stack.bottom = 0;
   redo_stack.top = redo_stack.bottom = 0;

   for (i = 0; i < EditSTACK_SIZE; i++)
   {
      if (edit_history[i].ins_text)
      {
	 XtFree(edit_history[i].ins_text);
	 edit_history[i].ins_text = NULL;
      }
      if (edit_history[i].del_text)
      {
	 XtFree(edit_history[i].del_text);
	 edit_history[i].del_text = NULL;
      }
   }

   changes_counter = 0;
   set_title_string();
}				/* reset_undo_redo */

#define  LAST_NO_ACTION   -1
#define  LAST_DELETE_SEP   0
#define  LAST_DELETE_CHAR  1
#define  LAST_INSERT_SEP   2
#define  LAST_INSERT_CHAR  3

static int LastAction = LAST_NO_ACTION;

void  clear_undo_redo(void)
{
   int      i;

   if (edit_history_initialized == False)
      return;

   for (i = 0; i < EditSTACK_SIZE; i++)
      edit_history[i].step = -1;

   LastAction = LAST_NO_ACTION;		/* undos after clear should not be merged ! */

   changes_counter = 0;
   set_title_string();
}				/* clear_undo_redo */

void  SaveActionForUndo(Widget w, XtPointer client_data, XtPointer call_data)
{
   /* save the current XmTextVerifyCallbackStruct for later use; additionally
      for delete operation save the deleted text in the undo_text->text->ptr;
      for an easy access in undo/redo I terminate undo_text->text->ptr with a
      '\0' (so it can be used straight away in the, for example, XmTextReplace).
    */
   XmTextVerifyPtr txt_data = (XmTextVerifyPtr)call_data;
   String   text_buffer;
   int      len;

   EditActionPtr undo_el;	/* txt_data for undo operation */

   int      i;
   Boolean  SepFlag;

   static XmTextPosition  LastInsert   = 0;
   static XmTextPosition  LastStartPos = 0;
   static XmTextPosition  LastEndPos   = 0;

   /* when you start editing allocate memory for text blocks in edit_history */
   if (!edit_history_initialized)
   {
      edit_history_initialized = True;
      for (i = 0; i < EditSTACK_SIZE; i++)
      {
	 edit_history[i].del_text = NULL;
	 edit_history[i].ins_text = NULL;
      }
   }

   /* the undo_el is a pointer to an element in edit_history, the
      element in edit_history has the same index as the top element on the
      undo_stack (which has a "cyclic" nature)
    */
   undo_el = &edit_history[undo_stack.top];

   /* if there was anything on the redo stack flush it (we are just about to
      put a new entry on the undo stack); make the redo button insensitive  */
   flush_edit_stack(&redo_stack);
   XtSetSensitive(redoMenuPB, False);

   /* for a modifying of the edit text allocate memory and the text;
      there may be two cases: text addition or deletion;  */
   len = txt_data->text->length;
   if (len > 0)   				/* text adding */
   {
      SepFlag = (strchr(WordSeps, txt_data->text->ptr[0]) == NULL) ? False : True;

      /* Collect multiple undos into one, if they are one character long, follow  */
      /* immediately after the last und and are of the same kind. This allows the */
      /* deletion of whole words with one undo instead of single characters.      */
      if ((undo_stack.top == undo_stack.bottom) ||
	  (len != 1) ||
	  (txt_data->currInsert != (LastInsert + 1)) ||
	  (txt_data->startPos != (LastEndPos + 1)) ||
	  ((SepFlag == True) && (LastAction != LAST_INSERT_SEP)) ||
	  ((SepFlag == False) && (LastAction != LAST_INSERT_CHAR)))
      {
	 /* can't use Xstrdup because ptr field is not NULL terminated */
	 undo_el->ins_text  = Xstrndup(txt_data->text->ptr, len);
	 undo_el->startPos  = txt_data->startPos;
	 undo_el->cursorPos = txt_data->currInsert;
      }
      else
      {
	 pop_edit_stack(&undo_stack, &undo_el, True);
	 i = strlen(undo_el->ins_text);
	 undo_el->ins_text = XtRealloc(undo_el->ins_text, (i + 2) * sizeof(char));
	 undo_el->ins_text[i] = txt_data->text->ptr[0];
	 undo_el->ins_text[i + 1] = '\0';
      }

      if (len == 1)
	 LastAction = (SepFlag == True) ? LAST_INSERT_SEP : LAST_INSERT_CHAR;
      else
	 LastAction = LAST_NO_ACTION;

      undo_el->endPos = txt_data->endPos;
   }

   /* now check if any data was deleted or not (if so store it) */
   len = (int)(txt_data->endPos - txt_data->startPos);	/* amount of text deleted */
   if (len > 0)
   {  /* text deleting ...(or just replacing with pending delete)
	 - store it but do not change text->length.
	 it will be used to recognize both those cases */
      /* get the text before deletion */
      text_buffer = XtMalloc((len + 1) * MB_CUR_MAX);
      XmTextGetSubstring(text_canvas, txt_data->startPos, len, (len + 1) * MB_CUR_MAX, text_buffer);
      SepFlag = (strchr(WordSeps, *text_buffer) == NULL) ? False : True;

      /* Condense multiple undos into one, if they are one character long, follow  */
      /* immediately after the last und and are of the same kind. This allows the  */
      /* reconstruction of whole words with one undo instead of single characters. */
      if ((undo_stack.top == undo_stack.bottom) ||
	  (len != 1) ||
	  ( (txt_data->currInsert != LastInsert) &&
	    (txt_data->currInsert != LastInsert - 1)) ||
	  ( (txt_data->endPos   != LastStartPos) &&
	    (txt_data->startPos != LastStartPos) ) ||
	  ((SepFlag == True) && (LastAction != LAST_DELETE_SEP)) ||
	  ((SepFlag == False) && (LastAction != LAST_DELETE_CHAR)))
      {
         undo_el->del_text  = text_buffer;
	 undo_el->endPos    = txt_data->endPos;
         undo_el->cursorPos = txt_data->currInsert;
      }
      else
      {
	 pop_edit_stack(&undo_stack, &undo_el, True);
	 i = strlen(undo_el->del_text);
	 undo_el->del_text = XtRealloc(undo_el->del_text, (i + 2) * sizeof(char));
	 if (txt_data->startPos != LastStartPos)
	 {
	    for (++i; i > 0; i--)
	       undo_el->del_text[i] = undo_el->del_text[i - 1];
	    undo_el->del_text[0] = *text_buffer;
	 }
	 else
	 {
	    undo_el->del_text[i] = *text_buffer;
	    undo_el->del_text[i+1] = '\0';
	 }
         XtFree(text_buffer);
      }

      if (len == 1)
	 LastAction = (SepFlag == True) ? LAST_DELETE_SEP : LAST_DELETE_CHAR;
      else
	 LastAction = LAST_NO_ACTION;

      undo_el->startPos = txt_data->startPos;
   }

   undo_el->step = 1;

   /* push the undo data on the undo stack ... */
   push_edit_stack(&undo_stack, undo_el, True);

   /* set sensitivity to the undo button */
   XtSetSensitive(undoMenuPB, True);

   LastInsert   = txt_data->currInsert;
   LastStartPos = txt_data->startPos;
   LastEndPos   = txt_data->endPos;
}				/* SaveActionForUndo */


/*****************************  UndoRedo  *************************************
**
**	Process undo or redo actions for text_canvas widget.
**      The kind of the action is set via action_type.
**      !!!!! The structure of the procedure is nearly the same as TextCB
**      so if you make changes there or here remeber to make the changes in
**      the other procedure. For full comments see TextCB !!!
*/
void     UndoRedo(int action_type)
{
   EditActionPtr txt_data = NULL;
   XmTextPosition start, end;

   long     ins_len, del_len;

   if (action_type == UNDO)
   {
      pop_edit_stack(&undo_stack, &txt_data, True);
      if (undo_stack.bottom == undo_stack.top)	/* last element was popped... */
      {
	 XtSetSensitive(undoMenuPB, False);
	 XFlush(XtDisplay(text_canvas));	/* sensitivity must be set before next keys are
						   processed, otherwise with a busy system we
						   could "skip over" the bottom of the stack
						   and all track would be lost !!! */
      }
   }
   else
   {
      pop_edit_stack(&redo_stack, &txt_data, False);
      if (redo_stack.bottom == redo_stack.top)	/* last element was popped... */
      {
	 XtSetSensitive(redoMenuPB, False);
	 XFlush(XtDisplay(text_canvas));	/* as above; be sure that the item is insensitive */
      }
   }

   if (txt_data == NULL)
      return;			/* should not happen, we are doing this just in
				   case. If it did happen it would mean that
				   the undo/redo was called when the appropriate
				   stack was empty, i.e. the sensitivity of the
				   appropriate button was set improperly (but now
				   would set correctly) */

   /* push the obtained element onto the OPPOSITE stack & set sensitivity */
   if (action_type == UNDO)
   {
      push_edit_stack(&redo_stack, txt_data, False);
      XtSetSensitive(redoMenuPB, True);
   }
   else
   {
      push_edit_stack(&undo_stack, txt_data, True);
      XtSetSensitive(undoMenuPB, True);
   }

   undo_redo_action = True;

   /* for the undo operation the action must be reversed !!!
      (what was stored on the undo stack is a plain copy of the
      operation performed, with only one addition for the delete
      operation, when the text was stored but the
      length field was not changed)
    */
   /* for SG there is a problem with a strlen function, if the
      argument is NULL we get the coredump !!! */
   if (txt_data->ins_text)
      ins_len = strlen(txt_data->ins_text);
   else
      ins_len = 0;
   if (txt_data->del_text)
      del_len = strlen(txt_data->del_text);
   else
      del_len = 0;

   if (action_type == UNDO)
   {
      start = txt_data->startPos;
      end   = start + ins_len;
      XmTextReplace(text_canvas, start, end, txt_data->del_text);
   }
   else			      /* redo */
   {
      start = txt_data->startPos;
      end = start + del_len;/*txt_data->endPos;*/
      XmTextReplace(text_canvas, start, end, txt_data->ins_text);
   }

   XmTextSetCursorPosition(text_canvas, txt_data->cursorPos);

   LastAction = LAST_NO_ACTION;
}				/* UndoRedo */
